/* reloc.c - contains code to parse the relocation information.
 * Written by Owen Klan - December 2002 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <elf.h>

#include "elfdump.h"

/* Local function prototypes */
static void dump_relocs(Elf32_Shdr *sec, Elf32_Ehdr *header, char *image);
static char * get_reloc_type(Elf32_Rel *rel);

void reloc_report(Elf32_Ehdr *header, char *image)
{
    Elf32_Shdr *sec;
    int i;
    
    sec = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    
    for (i = 0; i < ehalf_swap(header->e_shnum); i++, sec++) {
        if (eword_swap(sec->sh_type) == SHT_REL) {
	    dump_relocs(sec, header, image);
	}
    }
}

/* Procedure that will parse relocation information found in
 * a given section */
static void dump_relocs(Elf32_Shdr *sec, Elf32_Ehdr *header, char *image)
{
    Elf32_Rel *rel;
    char *type_str;
    int i;
    int num_rels;
    
    rel = (Elf32_Rel *)(image + eword_swap(sec->sh_offset));
    num_rels = (eword_swap(sec->sh_size) / eword_swap(sec->sh_entsize));
    
    fprintf(stdout, "\nRelocation Information (No addends) from %s\n",
	    (name_table + eword_swap(sec->sh_name)));
    fprintf(stdout, " Item  Offset    Symbol (Value)   Type\n");
    
    for (i = 0; i < num_rels; i++, rel++) {
	type_str = get_reloc_type(rel);
	fprintf(stdout, " %4d  %08X     %2d  %08X  %s\n",
		i, rel->r_offset,
		ELF32_R_SYM(eword_swap(rel->r_info)),
		get_dynsym_value(header, image,
				 ELF32_R_SYM(eword_swap(rel->r_info))),
		type_str);
	
	if (type_str)
	  free(type_str);
    }
}

/* Procedure that returns a string representing the type of relocation
 * described */
static char *get_reloc_type(Elf32_Rel *rel)
{
    char *rel_str = NULL;
    int rel_type = 0;
    
    if (!rel)
      return NULL;
    
    rel_type = ELF32_R_TYPE(eword_swap(rel->r_info));
    
    if ((rel_str = malloc(32)) == NULL) {
	return NULL;
    }
    
    switch (rel_type) {
#ifdef R_386_GOT32
    case R_386_GOT32:    sprintf(rel_str, "R_386_GOT32"); break;
#endif
#ifdef R_386_PLT32
    case R_386_PLT32:    sprintf(rel_str, "R_386_GOT32"); break;
#endif
#ifdef R_386_COPY
    case R_386_COPY:     sprintf(rel_str, "R_386_COPY"); break;
#endif
#ifdef R_386_GLOB_DAT
    case R_386_GLOB_DAT: sprintf(rel_str, "R_386_GLOB_DAT"); break;
#endif
#ifdef R_386_JMP_SLOT
    case R_386_JMP_SLOT: sprintf(rel_str, "R_386_JMP_SLOT"); break;
#endif
#ifdef R_386_RELATIVE
    case R_386_RELATIVE: sprintf(rel_str, "R_386_RELATIVE"); break;
#endif
#ifdef R_386_GOTOFF
    case R_386_GOTOFF:   sprintf(rel_str, "R_386_GOTOFF"); break;
#endif
#ifdef R_386_GOTPC
    case R_386_GOTPC:    sprintf(rel_str, "R_386_GOTPC"); break;
#endif
    default:
	sprintf(rel_str, "Unknown: %d\n", rel_type);
    }

    return rel_str;
}
